/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                        LibMSDOS.c                       **/
/**                                                         **/
/** This file contains MSDOS routines to work with VGA,     **/
/** joystick, and other hardware, including read/write GIF  **/
/** files using GIFLIB.                                     **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996-1999                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/
#ifdef MSDOS

#include "LibMSDOS.h"

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys\stat.h>
#include <i86.h>

#ifdef GIFLIB
#include <gif_lib.h>
#endif

#define NEXT(N) ((N+1)%MAXALLOC)

static int SyncScreen = 0; /* 1: Sync PutImage() to VBlanks  */
static int VGAMode    = 0; /* Last set screen mode           */

static pixel ColorAlloc[MAXALLOC];
static int ColorValue[256];
static int ColorCount[256];
static int AllocLo   = 0;
static int AllocHi   = 1;
static int MaxColors = 256;

/** WaitVBlank() *********************************************/
/** Wait VGA VBlank.                                        **/
/*************************************************************/
static void WaitVBlank(void)
{
  while(inp(0x3DA)&0x08);
  while(!(inp(0x3DA)&0x08));
}

/** ResetColors() ********************************************/
/** Reset color allocation table and set MaxColors.         **/
/*************************************************************/
static void ResetColors(int Max)
{
  int J;

  _disable();
  AllocLo=0;
  AllocHi=1;
  MaxColors=Max;
  outp(0x3C8,0);
  for(J=0;J<Max;J++)
  {
    ColorValue[J]=ColorCount[J]=0;
    outp(0x3C9,0);outp(0x3C9,0);outp(0x3C9,0);
  }
  _enable();
}

/** VGAText80() **********************************************/
/** Set 80x25 text screenmode.                              **/
/*************************************************************/
void VGAText80(void)
{
  union REGS R;
  R.w.ax=0x0003;int386(0x10,&R,&R);
  VGAMode=0;
}

/** VGA320x200() *********************************************/
/** Set 320x200x256 screenmode.                             **/
/*************************************************************/
void VGA320x200(void)
{
  union REGS R;
  R.w.ax=0x0013;int386(0x10,&R,&R);
  memset((char *)0xA0000,0,320*200);
  ResetColors(256);
  VGAMode=1;
}

/** VGA256x240() *********************************************/
/** Set 256x240x256 screenmode.                             **/
/*************************************************************/
void VGA256x240(void)
{
  static unsigned char Regs[24] =
  {
    0x4F,0x3F,0x40,0x92,
    0x44,0x10,0x0A,0x3E,
    0x00,0x41,0x00,0x00,
    0x00,0x00,0x00,0x00,
    0xEA,0xAC,0xDF,0x20,
    0x40,0xE7,0x06,0xA3
  };
  union REGS R;
  int J;

  /* Set 320x200x256 first */
  R.w.ax=0x0013;int386(0x10,&R,&R);

  _disable();                /* Interrupts off        */
  WaitVBlank();              /* Wait for VBlank       */
  outpw(0x3C4,0x100);        /* Sequencer reset       */
  outp(0x3C2,0xE3);          /* Misc. output register */
  outpw(0x3C4,0x300);        /* Clear sequencer reset */

  /* Unprotect registers 0-7 */
  outpw(0x3D4,((Regs[0x11]&0x7F)<<8)+0x11);
  for(J=0;J<24;J++) outpw(0x3D4,(Regs[J]<<8)+J);

  _enable();
  memset((char *)0xA0000,0,256*240);
  ResetColors(256);
  VGAMode=2;
}

/** VGASyncScreen() ******************************************/
/** Sync/Don't sync to VGA VBlanks.                         **/
/*************************************************************/
void VGASetSync(int Switch) { SyncScreen=Switch; }

/** VGANewImage() ********************************************/
/** Create a new image buffer.                              **/
/*************************************************************/
int VGANewImage(Image *Img,int W,int H)
{
  if(Img->Data=(char *)malloc(W*H))
  { memset(Img->Data,0,W*H);Img->W=W;Img->H=H;return(1); }
  Img->W=Img->H=0;return(0);
}

/** VGAFreeImage() *******************************************/
/** Free image buffer.                                      **/
/*************************************************************/
void VGAFreeImage(Image *Img)
{
  if(Img->Data) free(Img->Data);
  Img->Data=NULL;Img->W=Img->H=0;
}

/** VGAPutImage() ********************************************/
/** Put an image on the screen.                             **/
/*************************************************************/
void VGAPutImage(Image *Img,int X,int Y,int W,int H)
{
  char *D,*S;

  /* If not in graphics mode, fall out */
  if(!VGAMode) return;

  /* If incorrect parameters, fall out */
  if((X+W>Img->W)||(Y+H>Img->H)) return;

  /* If required, wait for VBlank */
  if(SyncScreen) WaitVBlank();

  /* Calculate destination address and copy data */
  if(VGAMode==2)
  {
    /* Clip image if bigger than screen */
    if(W>=256) { X+=(W-256)/2;W=256; }
    if(H>=240) { Y+=(H-240)/2;H=240; }
    /* Calculate source/destination */
    D=(char *)0xA0000+256*(240-H)/2+(256-W)/2;
    S=Img->Data+Img->W*Y+X;
    /* Copy image */
    for(;H;H--,D+=256,S+=Img->W) memcpy(D,S,W);
  }
  else
  {
    /* Clip image if bigger than screen */
    if(W>=320) { X+=(W-320)/2;W=320; }
    if(H>=200) { Y+=(H-200)/2;H=200; }
    /* Calculate source/destination */
    D=(char *)0xA0000+320*(200-H)/2+(320-W)/2;
    S=Img->Data+Img->W*Y+X;
    /* Copy image */
    for(;H;H--,D+=320,S+=Img->W) memcpy(D,S,W);
  }
}

/** VGAGetColor() ********************************************/
/** Allocate a new color.                                   **/
/*************************************************************/
pixel VGAGetColor(pixel R,pixel G,pixel B)
{
  register int I,J,K;

  R>>=2;G>>=2;B>>=2;
  I=((int)R<<16)|((int)G<<8)|B;

  /* If color allocation table full, flush its first third */
  if(AllocHi==AllocLo) VGAFlushColors(MAXALLOC*2/3);

  /* First, find if there is a same color allocated */
  for(J=1;J<MaxColors;J++)
    if(I==ColorValue[J]) { ColorCount[J]++;break; }

  /* If failed to match a color, look for a free entry */
  if(J==MaxColors)
    for(J=1;J<MaxColors;J++)
      if(!ColorCount[J]) break;

  /* If no free entries, flush allocation table */
  if(J==MaxColors)
  {
    for(AllocLo=NEXT(AllocLo);AllocLo!=AllocHi;AllocLo=NEXT(AllocLo))
      if(!--ColorCount[ColorAlloc[AllocLo]])
      { J=ColorAlloc[AllocLo];break; }

    if(AllocLo==AllocHi) { AllocHi=NEXT(AllocHi);return(0); }
  }

  /* If unused color, set it */
  if(!ColorCount[J])
  {
    ColorValue[J]=I;
    ColorCount[J]=1;
    _disable();
    outp(0x3C8,J);
    outp(0x3C9,R);
    outp(0x3C9,G);
    outp(0x3C9,B);
    _enable();
  }

  /* Add newly allocated color to the table */
  ColorAlloc[AllocHi]=J;
  AllocHi=NEXT(AllocHi);

  /* Success */
  return(J);
}

/** VGAFlushColors() *****************************************/
/** Flush allocated colors, but preserve N last ones.       **/
/*************************************************************/
void VGAFlushColors(int N)
{
  register int I,J;

  /* Number of colors to flush */
  I=AllocHi-AllocLo-1-N;
  if(AllocHi<=AllocLo) I+=MAXALLOC;

  /* Decrease usage counters */
  if(I>0)
    for(AllocLo=NEXT(AllocLo);I;AllocLo=NEXT(AllocLo))
    { ColorCount[ColorAlloc[AllocLo]]--;I--; }
}

/** GetJoyPos() **********************************************/
/** Get joystick positions.                                 **/
/*************************************************************/
int GetJoyPos(JoyPos *Pos,int Sticks)
{
  int J,I,D[4],Mask;

  Mask=(Sticks&STICK1? 0x03:0x00)|(Sticks&STICK2? 0x0C:0x00);
  D[0]=D[1]=D[2]=D[3]=-1;

  _disable();
  /* Wait for joystick port to stabilize */
  for(J=0;(inp(0x201)&Mask)&&(J<32767);J++);
  outp(0x201,0);
  for(J=0;Mask&&(J<32767);J++)
  {
    I=inp(0x201)&15;
    if(I!=Mask)
    {
      if((Mask^I)&1) D[0]=J;
      if((Mask^I)&2) D[1]=J;
      if((Mask^I)&4) D[2]=J;
      if((Mask^I)&8) D[3]=J;
      Mask&=I;
    }
  }
  I=~inp(0x201);
  _enable();

  J=0;
  if((D[0]>0)&&(D[1]>0))
  {
    Pos->X[0]=D[0];
    Pos->Y[0]=D[1];
    Pos->Fire[0]=(I>>4)&0x03;
    J|=STICK1;
  }
  if((D[0]>0)&&(D[1]>0))
  {
    Pos->X[1]=D[2];
    Pos->Y[1]=D[3];
    Pos->Fire[1]=(I>>6)&0x03;
    J|=STICK2;
  }
  return(J);
}

/** GetPath() ************************************************/
/** Extract pathname from filename. Returns pointer to the  **/
/** internal buffer containing the pathname.                **/
/*************************************************************/
char *GetPath(char *Name)
{
  static char Path[256];
  char *P,*T,*R;

  P=strrchr(Name,'\\');

  /* If path not found or too long, assume current */
  if(!P||(P-Name>200)) { strcpy(Path,"");return(Path); }

  /* Copy and return the pathname */
  for(T=Path;Name<=P;*T++=*Name++);
  *T='\0';return(Path);
}

/** NewFile() ************************************************/
/** Generate a new file name. Returns pointer to the        **/
/** internal buffer containing a new file name NAMEnnnn.EXT **/
/** where nnnn=0000..9999.                                  **/
/*************************************************************/
char *NewFile(char *File)
{
  static char Name[256];
  struct stat Status;
  char S[256],*P,*T;
  int J;

  /* If too long name, fall out */
  if(strlen(File)>200) { strcpy(Name,"");return(Name); }

  /* Make up the format string */
  for(T=S,P=File;*P&&(*P!='.');) *T++=*P++;
  *T='\0';strcat(S,"%04d");strcat(S,P);

  /* Scan through the filenames */
  for(J=0;J<10000;J++)
  {
    sprintf(Name,S,J);
    if(stat(Name,&Status)) break;
  }

  if(J==10000) strcpy(Name,"");
  return(Name);
}


/*************************************************************/
/** The following part should be compiled with GIFLIB.      **/
/*************************************************************/
#ifndef GIFLIB
int LoadGIF(char *File) { return(0); }
int SaveGIF(char *File) { return(0); }
#else /* GIFLIB */

/** LoadGIF() ************************************************/
/** Read picture from a .GIF file to screen. Returns the    **/
/** number of allocated colors or 0 if failed.              **/
/*************************************************************/
int LoadGIF(char *File)
{
  static int InterlacedOffset[] = { 0, 4, 2, 1 };
  static int InterlacedJumps[]  = { 8, 8, 4, 2 };

  GifFileType *D;
  GifRecordType RecordType;
  GifByteType *Extension;
  GifPixelType *Buf;
  GifColorType *Colors;
  int I,J,K,H,W,Lines,ExtCode,Width,Height;
  char *P;

  /* If not in the graphics mode, fall out */
  if(!VGAMode) return(0);
  /* If can't open .GIF file, fall out */
  if(!(D=DGifOpenFileName(File))) return(0);

  for(K=1,Lines=0;K&&DGifGetRecordType(D,&RecordType);)
    switch(RecordType)
    {
      case TERMINATE_RECORD_TYPE:
        K=0;break;
      case EXTENSION_RECORD_TYPE:
        if(DGifGetExtension(D,&ExtCode,&Extension))
          while(Extension)
            if(!DGifGetExtensionNext(D,&Extension)) break;
        break;
      case IMAGE_DESC_RECORD_TYPE:
        /* Get image description */
        if(!DGifGetImageDesc(D)) break;
        /* .GIF file must not be interlaced */
        if(D->IInterlace) break;
        /* Get image width/height */
        Width=D->IWidth? D->IWidth:D->SWidth;
        Height=D->IHeight? D->IHeight:D->SHeight;
        /* Get screen start/width/height */
        if(VGAMode==2) { P=(char *)0xA0000;W=256,H=240; }
        else           { P=(char *)0xA0000;W=320,H=200; }
        /* Allocate buffer and read the lines */
        if(Buf=(GifPixelType *)malloc(Width))
        {
          /* Skip extra lines */
          if(H>Height) P+=W*(H-Height)/2;
          else
            for(I=(Height-H)/2;I>=0;I--)
              if(!DGifGetLine(D,Buf,Width)) break;
          /* Read needed number of lines */
          for(Lines=0;Lines<H;Lines++,P+=W)
            if(!DGifGetLine(D,Buf,Width)) break;
            else
            {
              if(W>Width) { I=(W-Width)/2;J=0;K=Width; }
              else        { J=(Width-W)/2;I=0;K=W; }
              for(;K;K--) P[I++]=255-Buf[J++];
            }
          K=1;free(Buf);
        }
        break;
    }

  /* Set VGA palette */
  if(!Lines) I=0;
  else
  {
    Colors=D->IColorMap? D->IColorMap:D->SColorMap;
    I=1<<(D->IColorMap? D->IBitsPerPixel:D->SBitsPerPixel);
    outp(0x3C8,0);
    for(K=0;K<I;K++)
    {
      outp(0x3C8,255-K);
      outp(0x3C9,Colors[K].Red>>2);
      outp(0x3C9,Colors[K].Green>>2);
      outp(0x3C9,Colors[K].Blue>>2);
    }
  }

  DGifCloseFile(D);
  ResetColors(256-I);
  return(I);
}

/** SaveGIF() ************************************************/
/** Save screen into a .GIF file. Returns the number of     **/
/** scanlines written or 0 if failed.                       **/
/*************************************************************/
int SaveGIF(char *File)
{
  GifFileType *D;
  GifColorType Colors[256];
  int J,W,H;
  char *P;

  /* If not in the graphics mode, fall out */
  if(!VGAMode) return(0);
  /* If can't open .GIF file, fall out */
  if(!(D=EGifOpenFileName(File,0))) return(0);

  /* Read the palette */
  outp(0x3C7,0);
  for(J=0;J<256;J++)
  {
    Colors[J].Red=(inp(0x3C9)&0x3F)<<2;
    Colors[J].Green=(inp(0x3C9)&0x3F)<<2;
    Colors[J].Blue=(inp(0x3C9)&0x3F)<<2;
  }

  J=0;
  if(VGAMode==2) { P=(char *)0xA0000;W=256;H=240; }
  else           { P=(char *)0xA0000;W=320;H=200; }
  if(EGifPutScreenDesc(D,W,H,8,0,8,&Colors[0]))
    if(EGifPutImageDesc(D,0,0,W,H,0,8,&Colors[0]))
        for(J=0;J<H;J++,P+=W)
          if(!EGifPutLine(D,P,W)) break;

  EGifCloseFile(D);
  return(J);
}

#endif /* GIFLIB */
#endif /* MSDOS  */