/** EMULib Emulation Library *********************************/
/**                                                         **/
/**                           MIDI.c                        **/
/**                                                         **/
/** This file contains functions to generate MIDI files.    **/
/**                                                         **/
/** Copyright (C) Marat Fayzullin 1996-1999                 **/
/**     You are not allowed to distribute this software     **/
/**     commercially. Please, notify me, if you make any    **/
/**     changes to this file.                               **/
/*************************************************************/

#include "MIDI.h"

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#define TIMECONST         300 /* MIDI time constant  */

#define MelodicProg       80  /* Square Lead Patch   */
#define WhiteNoiseProg    126 /* Applause Patch      */
#define PeriodicNoiseProg 125 /* Helicopter Patch    */

typedef unsigned char byte;
typedef unsigned short word;

static int XVolume[16];
static int XFreq[16];
static int XType[16];
static byte XNote[16];
static word XPitch[16];
static byte XLevel[16];
static int Counter,XChannels;
static FILE *MIDIOut;

static struct { byte Note;word Wheel; } Freqs[4096] =
{
#include "MIDIFreq.h"
};

static void WriteDelta(void);
static void MidiMessage(byte D0,byte D1,byte D2);
static void NoteOff(byte Channel);
static void NoteOn(byte Channel,byte Note);
static void ChangePitch(byte Channel,word Wheel);
static void ChangeVolume(byte Channel,byte Volume);
static void WriteTempo(int Freq);

/** MIDITicks() **********************************************/
/** Increase number of ticks passed since the last command. **/
/*************************************************************/
void MIDITicks(int N) { Counter+=N; }

/** MIDISetVolume() ******************************************/
/** Set volume (0..255) for a given channel.                **/
/*************************************************************/
void MIDISetVolume(int Channel,int Volume)
{ MIDISound(Channel,XFreq[Channel],Volume); }

/** MIDISetFreq() ********************************************/
/** Set frequency (Hz) for a given channel.                 **/
/*************************************************************/
void MIDISetFreq(int Channel,int Freq)
{ MIDISound(Channel,Freq,XVolume[Channel]); }

/** MIDISound() **********************************************/
/** Set sound frequency (Hz) and volume (0..255) for a      **/
/** given channel.                                          **/
/*************************************************************/
void MIDISound(int Channel,int Freq,int Volume)
{
  int MIDIVolume,MIDINote;

  if((Freq<MIN_FREQ)||(Freq>MAX_FREQ)) Freq=0;
  MIDIVolume=(Volume&0xFF)/2;
  MIDINote=Freqs[Freq/3].Note;

  if(!Volume||!Freq) NoteOff(Channel);
  else
  {
    if(MIDINote!=XNote[Channel]) NoteOff(Channel);
    ChangePitch(Channel,Freqs[Freq/3].Wheel);
    ChangeVolume(Channel,MIDIVolume);
    if(!XNote[Channel]) NoteOn(Channel,MIDINote);
  }

  XVolume[Channel]=Volume;
  XFreq[Channel]=Freq;
}

/** InitMIDI() ***********************************************/
/** Write out MIDI header and initialize the variables. If  **/
/** failed, returns 0.                                      **/
/*************************************************************/
int InitMIDI(FILE *File,int Channels,int Ticks)
{
  static char MThd[] = "MThd\0\0\0\006\0\0\0\1";
                     /* ID  DataLen   Fmt Trks */
  static char MTrk[] = "MTrk\0\0\0\0";
                     /* ID  TrkLen   */
  int J;

  MIDIOut=File;
  if(fwrite(MThd,1,12,MIDIOut)!=12) return(0);
  fputc((TIMECONST>>8)&0xFF,MIDIOut);
  fputc(TIMECONST&0xFF,MIDIOut);
  if(fwrite(MTrk,1,8,MIDIOut)!=8) return(0);

  /* Clear the storage arrays and counter */
  for(J=0;J<16;J++) XNote[J]=XPitch[J]=XVolume[J]=XFreq[J]=XType[J]=0;
  Counter=0;XChannels=Channels;

  /* Write out the tempo */
  WriteTempo(Ticks);
  /* Put a MIDI controller 0 change in the file */
  /* to change to user bank 1                   */
  for(J=0;J<Channels;J++) MidiMessage(0xB0+J,0,1);
  /* Initially, load melodic patches */
  for(J=0;J<Channels;J++) MIDISetSound(J,SND_MELODIC);
  /* Set master volume */
  for(J=0;J<Channels;J++) MidiMessage(0xB0+J,7,0);

  return(1);
}

/** TrashMIDI() **********************************************/
/** Write out the footer and the file size. User must close **/
/** the file after calling this function.                   **/
/*************************************************************/
void TrashMIDI(void)
{
  long Length;
  int J;

  /* Turn sound off */
  for(J=0;J<XChannels;J++) NoteOff(J);
  /* End of track */
  MidiMessage(0xFF,0x2F,0x00);
  /* Put track length in file */
  fseek(MIDIOut,0,SEEK_END);
  Length=ftell(MIDIOut)-22;
  fseek(MIDIOut,18,SEEK_SET);
  fputc((Length>>24)&0xFF,MIDIOut);
  fputc((Length>>16)&0xFF,MIDIOut);
  fputc((Length>>8)&0xFF,MIDIOut);
  fputc(Length&0xFF,MIDIOut);
}

/** MIDISetSound() *******************************************/
/** Set sound type for a given channel.                     **/
/*************************************************************/
void MIDISetSound(int Channel,int Type)
{
  if(XType[Channel]!=Type)
  {
    XType[Channel]=Type;
    switch(Type)
    {
      case SND_MELODIC:  Type=MelodicProg;break;
      case SND_NOISE:    Type=WhiteNoiseProg;break;
      case SND_PERIODIC: Type=PeriodicNoiseProg;break;
      default:           return;
    }
    MidiMessage(0xC0+Channel,Type,255);
  }
}

/** WriteDelta() *********************************************/
/** Write number of ticks since the last MIDI command and   **/
/** reset the counter.                                      **/
/*************************************************************/
void WriteDelta(void)
{
  if(Counter<128) fputc(Counter,MIDIOut);
  else
    if(Counter<128*128)
    {
      fputc((Counter>>7)|0x80,MIDIOut);
      fputc(Counter&0x7F,MIDIOut);
    }
    else
    {
      fputc(((Counter>>14)&0x7F)|0x80,MIDIOut);
      fputc(((Counter>>7)&0x7F)|0x80,MIDIOut);
      fputc(Counter&0x7F,MIDIOut);
    }

  Counter=0;
}

/** MidiMessage() ********************************************/
/** Write out a MIDI message.                               **/
/*************************************************************/
void MidiMessage(byte D0,byte D1,byte D2)
{
 static byte Status=0;

 WriteDelta();
 if(D0!=Status) fputc(Status=D0,MIDIOut);
 if(D1<128)
 {
   fputc(D1,MIDIOut);
   if(D2<128) fputc(D2,MIDIOut);
 }
 D0&=0xF0;
 if((D0==0xC0)||(D0==0xD0)||(D0==0xF0)) Status=0;
}

/** NoteOff() ************************************************/
/** Turn off a note on a given channel.                     **/
/*************************************************************/
void NoteOff(byte Channel)
{
  if(XNote[Channel])
  {
    MidiMessage(0x80+Channel,XNote[Channel],127);
    XNote[Channel]=0;
  }
}

/** NoteOn() *************************************************/
/** Turn on a note on a given channel.                      **/
/*************************************************************/
void NoteOn(byte Channel,byte Note)
{
  if(XNote[Channel]!=Note)
  {
    if(XNote[Channel]) NoteOff(Channel);
    MidiMessage(0x90+Channel,Note,127);
    XNote[Channel]=Note;
  }
}

/** ChangePitch() ********************************************/
/** Change pitch bend on a channel.                         **/
/*************************************************************/
void ChangePitch(byte Channel,word Wheel)
{
  if(XPitch[Channel]!=Wheel)
  {
    MidiMessage(0xE0+Channel,Wheel&0x7F,(Wheel>>7)&0x7F);
    XPitch[Channel]=Wheel;
  }
}

/** ChangeVolume() *******************************************/
/** Change volume on a channel.                             **/
/*************************************************************/
void ChangeVolume(byte Channel,byte Volume)
{
  if(XLevel[Channel]!=Volume)
  {
    MidiMessage(0xB0+Channel,7,Volume);
    XLevel[Channel]=Volume;
  }
}

/** WriteTempo() *********************************************/
/** Write out soundtrack tempo (Hz).                        **/
/*************************************************************/
void WriteTempo(int Freq)
{
  int J;

  J=500000*TIMECONST*2/Freq;
  WriteDelta();
  fputc(0xFF,MIDIOut);
  fputc(0x51,MIDIOut);
  fputc(0x03,MIDIOut);
  fputc((J>>16)&0xFF,MIDIOut);
  fputc((J>>8)&0xFF,MIDIOut);
  fputc(J&0xFF,MIDIOut);
}

